from math import floor
from pathlib import Path

import dask.array as da
from numpy import ascontiguousarray, empty, float32, float64, nan_to_num, uint8, uint64
from tqdm import tqdm
from xarray import DataArray

__all__ = ["write_bed"]


def write_bed(filepath: Path, X: DataArray, major: str, verbose: bool):
    """
    Write BED file.

    It assumes that ``X`` is a sample-by-variant matrix.
    """
    from .bed_reader import lib

    G = da.asanyarray(X)

    if major == "variant":
        G = G.T

    major_code = 1 if major == "variant" else 0
    e = lib.write_bed_header(str(filepath).encode(), major_code)
    if e != 0:
        raise RuntimeError(f"Failure while writing BED file {filepath}.")

    nrows = G.shape[0]
    ncols = G.shape[1]

    row_chunk = max(1, floor((1024 * 1024 * 256) / ncols))
    row_chunk = min(row_chunk, nrows)

    G = G.rechunk((row_chunk, ncols))

    row_start = 0
    for chunk in tqdm(G.chunks[0], "Writing BED", disable=not verbose):
        data = G[row_start : row_start + chunk, :].compute()
        if data.dtype not in [float32, float64]:
            msg = "Unsupported data type. "
            msg += "Please, provide a dosage matrix in either "
            msg += "float32 or float64 format."
            raise ValueError(msg)

        _write_bed_chunk(filepath, data)
        row_start += chunk


def _write_bed_chunk(filepath: Path, X):
    from .bed_reader import ffi, lib

    base_type = uint8
    base_size = base_type().nbytes
    base_repr = "uint8_t"

    nan_to_num(X, False, 3.0)
    G = ascontiguousarray(X, base_type)
    assert G.flags.aligned

    strides = empty(2, uint64)
    strides[:] = G.strides
    strides //= base_size

    e = lib.write_bed_chunk(
        str(filepath).encode(),
        G.shape[1],
        G.shape[0],
        ffi.cast(f"{base_repr} *", G.ctypes.data),
        ffi.cast("uint64_t *", strides.ctypes.data),
    )
    if e != 0:
        raise RuntimeError(f"Failure while writing BED file {filepath}.")



from pathlib import Path
from typing import Optional, Union

from numpy import arange, float64, full, int32
from xarray import DataArray

from ._bed_write import write_bed

__all__ = ["write_plink1_bin"]


def write_plink1_bin(
    G: DataArray,
    bed: Union[str, Path],
    bim: Optional[Union[str, Path]] = None,
    fam: Optional[Union[str, Path]] = None,
    major: str = "variant",
    verbose: bool = True,
):
    """
    Write PLINK 1 binary files into a data array.

    A PLINK 1 binary file set consists of three files:

    - BED: containing the genotype.
    - BIM: containing variant information.
    - FAM: containing sample information.

    The user must provide the genotype (dosage) via a :class:`xarray.DataArray` matrix
    with data type :const:`numpy.float32` or :const:`numpy.float64`. That matrix must
    have two named dimensions: **sample** and **variant**. The only allowed values for
    the genotype are: :const:`0`, :const:`1`, :const:`2`, and :data:`math.nan`.

    Examples
    --------
    .. testsetup:

        >>> import os
        >>> import shutil
        >>> import tempfile
        >>>
        >>> old_path = os.getcwd()
        >>> tmp_path = tempfile.mkdtemp()
        >>> os.chdir(tmp_path)

    .. doctest::

        >>> from xarray import DataArray
        >>> from pandas_plink import read_plink1_bin, write_plink1_bin
        >>>
        >>> G = DataArray([[0.0]], dims=["sample", "variant"])
        >>> write_plink1_bin(G, "sample.bed", verbose=False)
        >>> print(G)
        <xarray.DataArray (sample: 1, variant: 1)>
        array([[0.00]])
        Dimensions without coordinates: sample, variant
        >>> G = read_plink1_bin("sample.bed", verbose=False)
        >>> print(G)
        <xarray.DataArray 'genotype' (sample: 1, variant: 1)>
        dask.array<transpose, shape=(1, 1), dtype=float32, chunksize=(1, 1), chunktype=numpy.ndarray>
        Coordinates:
          * sample   (sample) object '0'
          * variant  (variant) <U8 'variant0'
            fid      (sample) object '0'
            iid      (sample) object '0'
            father   (sample) object '?'
            mother   (sample) object '?'
            gender   (sample) object '0'
            trait    (sample) object '-9'
            chrom    (variant) object '?'
            snp      (variant) object '0'
            cm       (variant) float64 0.0
            pos      (variant) int32 0
            a0       (variant) object '?'
            a1       (variant) object '?'

    The following example reads two BED files and two BIM files correspondig to
    chromosomes 11 and 12, and read a single FAM file whose filename is inferred from
    the BED filenames. It then saves the resulting matrix to disk.

    .. doctest::

        >>> from os.path import join
        >>> from pandas_plink import read_plink1_bin, write_plink1_bin
        >>> from pandas_plink import get_data_folder
        >>>
        >>> G = read_plink1_bin(join(get_data_folder(), "chr*.bed"), verbose=False)
        >>> write_plink1_bin(G, "all.bed", verbose=False)
        >>> G = read_plink1_bin("all.bed", verbose=False)
        >>> print(G)
        <xarray.DataArray 'genotype' (sample: 14, variant: 1252)>
        dask.array<transpose, shape=(14, 1252), dtype=float32, chunksize=(14, 1024), chunktype=numpy.ndarray>
        Coordinates:
          * sample   (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
          * variant  (variant) <U11 'variant0' 'variant1' ... 'variant1251'
            fid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
            iid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
            father   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            mother   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            gender   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            trait    (sample) object '-9' '-9' '-9' '-9' '-9' ... '-9' '-9' '-9' '-9'
            chrom    (variant) object '11' '11' '11' '11' '11' ... '12' '12' '12' '12'
            snp      (variant) object '316849996' '316874359' ... '373081507'
            cm       (variant) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
            pos      (variant) int32 157439 181802 248969 ... 27163741 27205125 27367844
            a0       (variant) object 'C' 'G' 'G' 'C' 'C' 'T' ... 'A' 'G' 'A' 'T' 'G'
            a1       (variant) object 'T' 'C' 'C' 'T' 'T' 'A' ... 'G' 'A' 'T' 'C' 'A'

    .. testcleanup::

        >>> os.chdir(old_path)
        >>> shutil.rmtree(tmp_path)

    Parameters
    ----------
    G
        Genotype matrix with metainformation about samples and variants.
    bed
        Path to a BED file.
    bim
        Path to a BIM file.It defaults to :const:`None`, in which case it will try to be
        inferred.
    fam
        Path to a FAM file. It defaults to :const:`None`, in which case it will try to
        be inferred.
    major
        It can be either :const:`"sample"` or :const:`"variant"` (recommended and
        default). Specify the matrix layout on the BED file.
    verbose
        :const:`True` for progress information; :const:`False` otherwise.
    """
    if G.ndim != 2:
        raise ValueError("G has to be bidimensional")

    if set(list(G.dims)) != set(["sample", "variant"]):
        raise ValueError("G has to have both `sample` and `variant` dimensions.")

    if major not in ["sample", "variant"]:
        raise ValueError("Major can be either `sample` or `variant`.")

    G = G.transpose("sample", "variant")

    bed = Path(bed)

    if bim is None:
        bim = bed.with_suffix(".bim")

    if fam is None:
        fam = bed.with_suffix(".fam")

    bim = Path(bim)
    fam = Path(fam)

    G = _fill_sample(G)
    G = _fill_variant(G)
    write_bed(bed, G, major, verbose)

    _echo("Writing FAM... ", end="", disable=not verbose)
    _write_fam(fam, G)
    _echo("done.", disable=not verbose)

    _echo("Writing BIM... ", end="", disable=not verbose)
    _write_bim(bim, G)
    _echo("done.", disable=not verbose)


def _echo(msg: str, end: str = "\n", disable: bool = False):
    if not disable:
        print(msg, end=end, flush=True)


def _fill_sample(G: DataArray) -> DataArray:
    nsamples = G.sample.shape[0]
    if "fid" not in G.sample.coords:
        G = G.assign_coords(fid=("sample", G.sample.values))

    if "iid" not in G.sample.coords:
        G = G.assign_coords(iid=("sample", G.sample.values))

    if "father" not in G.sample.coords:
        G = G.assign_coords(father=("sample", full(nsamples, "?", object)))

    if "mother" not in G.sample.coords:
        G = G.assign_coords(mother=("sample", full(nsamples, "?", object)))

    if "gender" not in G.sample.coords:
        G = G.assign_coords(gender=("sample", full(nsamples, "0", object)))

    if "trait" not in G.sample.coords:
        G = G.assign_coords(trait=("sample", full(nsamples, "-9", object)))
    return G


def _fill_variant(G: DataArray) -> DataArray:
    nvariants = G.variant.shape[0]
    if "chrom" not in G.variant.coords:
        G = G.assign_coords(chrom=("variant", full(nvariants, "?", object)))

    if "snp" not in G.variant.coords:
        G = G.assign_coords(snp=("variant", arange(nvariants, dtype=object)))

    if "cm" not in G.variant.coords:
        G = G.assign_coords(cm=("variant", full(nvariants, 0.0, float64)))

    if "pos" not in G.variant.coords:
        G = G.assign_coords(pos=("variant", full(nvariants, 0.0, int32)))

    if "a0" not in G.variant.coords:
        G = G.assign_coords(a0=("variant", full(nvariants, "?", object)))

    if "a1" not in G.variant.coords:
        G = G.assign_coords(a1=("variant", full(nvariants, "?", object)))
    return G


def _write_fam(filepath: Path, G: DataArray):
    from pandas import DataFrame

    df = DataFrame()
    cols = [
        ("fid", object),
        ("iid", object),
        ("father", object),
        ("mother", object),
        ("gender", object),
        ("trait", object),
    ]

    for col, col_type in cols:
        df[col] = G.sample[col].values.astype(col_type)

    df.to_csv(
        filepath,
        index=False,
        sep="\t",
        header=None,
        encoding="ascii",
        line_terminator="\n",
    )


def _write_bim(filepath: Path, G: DataArray):
    from pandas import DataFrame

    df = DataFrame()
    cols = [
        ("chrom", object),
        ("snp", object),
        ("cm", float64),
        ("pos", int32),
        ("a0", object),
        ("a1", object),
    ]

    for col, col_type in cols:
        df[col] = G.variant[col].values.astype(col_type)

    df.to_csv(
        filepath,
        index=False,
        sep="\t",
        header=None,
        encoding="ascii",
        line_terminator="\n",
    )



from os.path import dirname, join, realpath

import pytest
from numpy import array, dtype, nan
from numpy.testing import assert_array_equal, assert_equal

from pandas_plink import example_file_prefix, read_plink, read_plink1_bin


def test_read_plink():

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")

    (bim, fam, bed) = read_plink(file_prefix, verbose=False)
    assert_equal(bed.dtype, dtype("float32"))

    assert_array_equal(bim.query("chrom=='1' and pos==72515")["snp"], ["rs4030300"])
    assert_array_equal(bim.query("chrom=='1'").shape, [10, 7])
    assert_array_equal(
        fam.query("fid=='Sample_2' and iid=='Sample_2'")["trait"], ["-9"]
    )

    assert_array_equal(
        bed,
        array(
            [
                [2, 2, 1],
                [2, 1, 2],
                [nan, nan, nan],
                [nan, nan, 1],
                [2, 2, 2],
                [2, 2, 2],
                [2, 1, 0],
                [2, 2, 2],
                [1, 2, 2],
                [2, 1, 2],
            ]
        ),
    )


def test_read_plink_prefix_dot():
    with pytest.raises(IOError):
        read_plink("/home/joao/84757.genotypes.norm.renamed")


def test_read_plink_wildcard():
    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "chr*")

    (bim, fam, bed) = read_plink(file_prefix, verbose=False)
    assert_array_equal(bim[bim["chrom"] == "11"]["i"].values[:2], [0, 1])
    assert_array_equal(bim[bim["chrom"] == "12"]["i"].values[:2], [779, 780])


def test_read_plink1_bin():

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False)
    assert_equal(G.data.dtype, dtype("float32"))

    snp = G.where((G.chrom == "1") & (G.pos == 72515), drop=True)["snp"].values
    assert_array_equal(snp, ["rs4030300"])

    shape = G.where(G.chrom == "1", drop=True).shape
    assert_array_equal(shape, [3, 10])

    shape = G.where(G.chrom == "2", drop=True).shape
    assert_array_equal(shape, [3, 0])

    g = G.where((G.fid == "Sample_2") & (G.iid == "Sample_2"), drop=True)
    assert_array_equal(g["trait"].values, ["-9"])

    arr = [
        [2.0, 2.0, nan, nan, 2.0, 2.0, 2.0, 2.0, 1.0, 2.0],
        [2.0, 1.0, nan, nan, 2.0, 2.0, 1.0, 2.0, 2.0, 1.0],
        [1.0, 2.0, nan, 1.0, 2.0, 2.0, 0.0, 2.0, 2.0, 2.0],
    ]
    assert_array_equal(G, arr)


def test_read_plink1_bin_a0():

    datafiles = join(dirname(realpath(__file__)), "data_files")
    file_prefix = join(datafiles, "data")
    bim = file_prefix + ".bim"
    bed = file_prefix + ".bed"
    fam = file_prefix + ".fam"

    G = read_plink1_bin(bed, bim, fam, verbose=False, ref="a0")
    assert_equal(G.data.dtype, dtype("float32"))

    snp = G.where((G.chrom == "1") & (G.pos == 72515), drop=True)["snp"].values
    assert_array_equal(snp, ["rs4030300"])

    shape = G.where(G.chrom == "1", drop=True).shape
    assert_array_equal(shape, [3, 10])

    shape = G.where(G.chrom == "2", drop=True).shape
    assert_array_equal(shape, [3, 0])

    g = G.where((G.fid == "Sample_2") & (G.iid == "Sample_2"), drop=True)
    assert_array_equal(g["trait"].values, ["-9"])

    arr = [
        [0.0, 0.0, nan, nan, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
        [0.0, 1.0, nan, nan, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0],
        [1.0, 0.0, nan, 1.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0],
    ]
    assert_array_equal(G, arr)


def test_read_plink1_bin_wildcard_not_found():
    datafiles = join(dirname(realpath(__file__)), "data_files")
    bed_files = join(datafiles, "chrr*.bed")

    with pytest.raises(ValueError):
        read_plink1_bin(bed_files, verbose=False)

    bed_files = join(datafiles, "chr*.bed")
    with pytest.raises(ValueError):
        read_plink1_bin(bed_files, "chr11.bim", verbose=False)

    bed_files = join(datafiles, "chr*.bed")
    bim_files = join(datafiles, "chrr*.bim")
    with pytest.raises(ValueError):
        read_plink1_bin(bed_files, bim_files, verbose=False)

    bed_files = join(datafiles, "chr*.bed")
    bim_files = join(datafiles, "chr*.bim")
    fam_files = join(datafiles, "chr*.fam")
    with pytest.warns(UserWarning):
        read_plink1_bin(bed_files, bim_files, fam_files, verbose=True)


def test_read_plink1_bin_wildcard():
    datafiles = join(dirname(realpath(__file__)), "data_files")
    bed_files = join(datafiles, "chr*.bed")

    G = read_plink1_bin(bed_files, verbose=False)
    G.where(G.chrom == "11", drop=True).values
    assert_equal(G.where(G.chrom == "11", drop=True).shape, (14, 779))
    assert_equal(G.where(G.chrom == "12", drop=True).shape, (14, 473))
    x = [[0.00, 0.00], [0.00, 1.00]]
    assert_equal(G.where(G.chrom == "11", drop=True).values[:2, :2], x)


def test_example_file_prefix():
    with pytest.warns(DeprecationWarning):
        example_file_prefix()



import warnings
from collections import OrderedDict as odict
from glob import glob
from os.path import basename, dirname, join
from typing import Optional

from pandas import StringDtype
from xarray import DataArray

from ._allele import Allele
from ._bed_read import read_bed
from ._chunk import Chunk
from ._util import last_replace

__all__ = ["read_plink", "read_plink1_bin"]


def read_plink(file_prefix, verbose=True):
    """
    Read PLINK files into data frames.

    .. note::

        The function :func:`pandas_plink.read_plink1_bin` provides an alternative
        interface to read the same files.

    The genotype values can be either :const:`0`, :const:`1`, :const:`2`, or
    :data:`math.nan`:

    - :const:`0` Homozygous having the first allele (given by coordinate **a0**)
    - :const:`1` Heterozygous
    - :const:`2` Homozygous having the second allele (given by coordinate **a1**)
    - :data:`math.nan` Missing genotype

    Examples
    --------
    The following example reads two BED files and two BIM files correspondig to
    chromosomes 11 and 12, and read a single FAM file whose filename is inferred from
    the BED filenames.

    .. doctest::

        >>> from os.path import join
        >>> from pandas_plink import read_plink
        >>> from pandas_plink import get_data_folder
        >>> (bim, fam, bed) = read_plink(join(get_data_folder(), "chr*.bed"),
        ...                              verbose=False)
        >>> print(bim.head())
          chrom        snp       cm     pos a0 a1  i
        0    11  316849996     0.00  157439  C  T  0
        1    11  316874359     0.00  181802  G  C  1
        2    11  316941526     0.00  248969  G  C  2
        3    11  317137620     0.00  445063  C  T  3
        4    11  317534352     0.00  841795  C  T  4
        >>> print(fam.head())
            fid   iid father mother gender trait  i
        0  B001  B001      0      0      0    -9  0
        1  B002  B002      0      0      0    -9  1
        2  B003  B003      0      0      0    -9  2
        3  B004  B004      0      0      0    -9  3
        4  B005  B005      0      0      0    -9  4
        >>> print(bed.compute())
        [[0.00 0.00 0.00 ... 2.00 2.00 0.00]
         [0.00 1.00 0.00 ... 2.00 1.00 0.00]
         [2.00 2.00 2.00 ... 0.00 0.00 2.00]
         ...
         [2.00 0.00 0.00 ... 2.00 2.00 1.00]
         [2.00 0.00 0.00 ... 2.00 2.00 0.00]
         [0.00  nan 0.00 ... 1.00 2.00 0.00]]

    Parameters
    ----------
    file_prefix : str
        Path prefix to the set of PLINK files. It supports loading many BED files at
        once using globstrings wildcard.
    verbose : bool
        :const:`True` for progress information; :const:`False` otherwise.

    Returns
    -------
    alleles : :class:`pandas.DataFrame`
        Alleles.
    samples : :class:`pandas.DataFrame`
        Samples.
    genotypes : :class:`dask.array.Array`
        Genotype.
    """
    import pandas as pd
    from dask.array import concatenate
    from tqdm import tqdm

    file_prefixes = sorted(glob(file_prefix))
    if len(file_prefixes) == 0:
        file_prefixes = [file_prefix.replace("*", "")]

    file_prefixes = sorted(_clean_prefixes(file_prefixes))

    fn = []
    for fp in file_prefixes:
        fn.append({s: "%s.%s" % (fp, s) for s in ["bed", "bim", "fam"]})

    pbar = tqdm(desc="Mapping files", total=3 * len(fn), disable=not verbose)

    bim = _read_file(fn, lambda fn: _read_bim(fn["bim"]), pbar)
    if len(file_prefixes) > 1:
        if verbose:
            msg = "Multiple files read in this order: {}"
            print(msg.format([basename(f) for f in file_prefixes]))

    nmarkers = dict()
    index_offset = 0
    for i, bi in enumerate(bim):
        nmarkers[fn[i]["bed"]] = bi.shape[0]
        bi["i"] += index_offset
        index_offset += bi.shape[0]
    bim = pd.concat(bim, axis=0, ignore_index=True)

    fam = _read_file([fn[0]], lambda fn: _read_fam(fn["fam"]), pbar)[0]
    nsamples = fam.shape[0]

    ref = Allele.a1
    bed = _read_file(
        fn,
        lambda f: _read_bed(f["bed"], nsamples, nmarkers[f["bed"]], ref, Chunk()),
        pbar,
    )

    bed = concatenate(bed, axis=0)

    pbar.close()

    return (bim, fam, bed)


def read_plink1_bin(
    bed: str,
    bim: Optional[str] = None,
    fam: Optional[str] = None,
    verbose: bool = True,
    ref: str = "a1",
    chunk: Chunk = Chunk(),
) -> DataArray:
    """
    Read PLINK 1 binary files [a]_ into a data array.

    A PLINK 1 binary file set consists of three files:

    - BED: containing the genotype.
    - BIM: containing variant information.
    - FAM: containing sample information.

    The user might provide a single file path to a BED file, from which this function
    will try to infer the file path of the other two files.
    This function also allows the user to provide file path to multiple BED and
    BIM files, as it is common to have a data set split into multiple files, one per
    chromosome.

    This function returns a samples-by-variants matrix. This is a special kind of matrix
    with rows and columns having multiple coordinates each. Those coordinates have the
    metainformation contained in the BIM and FAM files.

    Examples
    --------
    The following example reads two BED files and two BIM files correspondig to
    chromosomes 11 and 12, and read a single FAM file whose filename is inferred from
    the BED filenames.

    .. doctest::

        >>> from os.path import join
        >>> from pandas_plink import read_plink1_bin
        >>> from pandas_plink import get_data_folder
        >>> G = read_plink1_bin(join(get_data_folder(), "chr*.bed"), verbose=False)
        >>> print(G)
        <xarray.DataArray 'genotype' (sample: 14, variant: 1252)>
        dask.array<concatenate, shape=(14, 1252), dtype=float32, chunksize=(14, 779), chunktype=numpy.ndarray>
        Coordinates:
          * sample   (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
          * variant  (variant) <U11 'variant0' 'variant1' ... 'variant1251'
            fid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
            iid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
            father   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            mother   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            gender   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            trait    (sample) object '-9' '-9' '-9' '-9' '-9' ... '-9' '-9' '-9' '-9'
            chrom    (variant) object '11' '11' '11' '11' '11' ... '12' '12' '12' '12'
            snp      (variant) object '316849996' '316874359' ... '373081507'
            cm       (variant) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
            pos      (variant) int32 157439 181802 248969 ... 27163741 27205125 27367844
            a0       (variant) object 'C' 'G' 'G' 'C' 'C' 'T' ... 'A' 'G' 'A' 'T' 'G'
            a1       (variant) object 'T' 'C' 'C' 'T' 'T' 'A' ... 'G' 'A' 'T' 'C' 'A'
        >>> print(G.shape)
        (14, 1252)

    Suppose we want the genotypes of the chromosome 11 only:

    .. doctest::

        >>> G = G.where(G.chrom == "11", drop=True)
        >>> print(G)
        <xarray.DataArray 'genotype' (sample: 14, variant: 779)>
        dask.array<where, shape=(14, 779), dtype=float32, chunksize=(14, 779), chunktype=numpy.ndarray>
        Coordinates:
          * sample   (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
          * variant  (variant) <U11 'variant0' 'variant1' ... 'variant777' 'variant778'
            fid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
            iid      (sample) object 'B001' 'B002' 'B003' ... 'B012' 'B013' 'B014'
            father   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            mother   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            gender   (sample) object '0' '0' '0' '0' '0' '0' ... '0' '0' '0' '0' '0' '0'
            trait    (sample) object '-9' '-9' '-9' '-9' '-9' ... '-9' '-9' '-9' '-9'
            chrom    (variant) object '11' '11' '11' '11' '11' ... '11' '11' '11' '11'
            snp      (variant) object '316849996' '316874359' ... '345698259'
            cm       (variant) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0
            pos      (variant) int32 157439 181802 248969 ... 28937375 28961091 29005702
            a0       (variant) object 'C' 'G' 'G' 'C' 'C' 'T' ... 'A' 'C' 'A' 'A' 'T'
            a1       (variant) object 'T' 'C' 'C' 'T' 'T' 'A' ... 'G' 'T' 'G' 'C' 'C'
        >>> print(G.shape)
        (14, 779)

    Lets now print the genotype value of the sample `B003` for variant `variant5`:

    .. doctest::

        >>> print(G.sel(sample="B003", variant="variant5").values)
        0.0

    The special matrix we return is of type :class:`xarray.DataArray`. More information
    about it can be found at the `xarray documentation <http://xarray.pydata.org>`_.


    Parameters
    ----------
    bed
        Path to a BED file. It can contain shell-style wildcards to indicate multiple
        BED files.
    bim
        Path to a BIM file. It can contain shell-style wildcards to indicate multiple
        BIM files. It defaults to :const:`None`, in which case it will try to be inferred.
    fam
        Path to a FAM file. It defaults to :const:`None`, in which case it will try to be
        inferred.
    verbose
        :const:`True` for progress information; :const:`False` otherwise.
    ref
        Reference allele. Specify which allele the dosage matrix will count. It can
        be either :const:`"a1"` (default) or :const:`"a0"`.
    chunk
        Data chunk specification. Useful to adjust the trade-off between computational
        overhead and IO usage. See :class:`pandas_plink.Chunk`.

    Returns
    -------
    :class:`xarray.DataArray`
        Genotype with metadata.

    References
    ----------
    .. [a] PLINK 1 binary. https://www.cog-genomics.org/plink/2.0/input#bed
    """
    import dask.array as da
    import pandas as pd
    from tqdm import tqdm

    bed_files = sorted(glob(bed))
    if len(bed_files) == 0:
        raise ValueError("No BED file has been found.")

    if bim is None:
        bim_files = [last_replace(f, ".bed", ".bim") for f in bed_files]
    else:
        bim_files = sorted(glob(bim))
    if len(bim_files) == 0:
        raise ValueError("No BIM file has been found.")

    if fam is None:
        fam_files = [last_replace(f, ".bed", ".fam") for f in bed_files]
    else:
        fam_files = sorted(glob(fam))
    if len(fam_files) == 0:
        raise ValueError("No FAM file has been found.")

    if len(bed_files) != len(bim_files):
        raise ValueError("The numbers of BED and BIM files must match.")

    if len(fam_files) > 1:
        msg = "More than one FAM file has been specified. Only the first one will be "
        msg += "considered."
        if verbose:
            warnings.warn(msg, UserWarning)
        fam_files = fam_files[:1]

    nfiles = len(bed_files) + len(bim_files) + 1
    pbar = tqdm(desc="Mapping files", total=nfiles, disable=not verbose)

    bims = _read_file(bim_files, lambda f: _read_bim_noi(f), pbar)
    nmarkers = {bed_files[i]: b.shape[0] for i, b in enumerate(bims)}
    bim_df = pd.concat(bims, axis=0, ignore_index=True)
    fam_df = _read_file(fam_files, lambda f: _read_fam_noi(f), pbar)[0]

    nsamples = fam_df.shape[0]
    sample_ids = fam_df["iid"]
    nvariants = bim_df.shape[0]
    variant_ids = [f"variant{i}" for i in range(nvariants)]

    if ref == "a1":
        ref_al = Allele.a1
    elif ref == "a0":
        ref_al = Allele.a0
    else:
        raise ValueError("Unknown reference allele.")

    G = _read_file(
        bed_files, lambda f: _read_bed(f, nsamples, nmarkers[f], ref_al, chunk).T, pbar
    )
    G = da.concatenate(G, axis=1)

    G = DataArray(G, dims=["sample", "variant"], coords=[sample_ids, variant_ids])
    sample = {c: ("sample", fam_df[c]) for c in fam_df.columns}
    variant = {c: ("variant", bim_df[c]) for c in bim_df.columns}
    G = G.assign_coords(**sample)
    G = G.assign_coords(**variant)
    G.name = "genotype"

    pbar.close()

    return G


def _read_file(fn, read_func, pbar):
    data = []
    for f in fn:
        data.append(read_func(f))
        pbar.update(1)
    return data


def _read_csv(fn, header):
    from pandas import read_csv

    return read_csv(
        fn,
        delim_whitespace=True,
        header=None,
        names=list(header.keys()),
        dtype=header,
        compression=None,
        engine="c",
    )


def _read_bim(fn):
    df = _read_bim_noi(fn)
    df["i"] = range(df.shape[0])
    return df


def _read_bim_noi(fn):
    from numpy import float64, int32

    header = odict(
        [
            ("chrom", StringDtype()),
            ("snp", StringDtype()),
            ("cm", float64),
            ("pos", int32),
            ("a0", StringDtype()),
            ("a1", StringDtype()),
        ]
    )
    return _read_csv(fn, header)


def _read_fam(fn):
    df = _read_fam_noi(fn)
    df["i"] = range(df.shape[0])
    return df


def _read_fam_noi(fn):
    header = odict(
        [
            ("fid", StringDtype()),
            ("iid", StringDtype()),
            ("father", StringDtype()),
            ("mother", StringDtype()),
            ("gender", StringDtype()),
            ("trait", StringDtype()),
        ]
    )

    return _read_csv(fn, header)


def _read_bed(fn, nsamples, nvariants, ref: Allele, chunk: Chunk):
    _check_bed_header(fn)
    major = _major_order(fn)

    ncols = nvariants
    nrows = nsamples
    row_chunk = nrows if chunk.nsamples is None else min(nrows, chunk.nsamples)
    col_chunk = ncols if chunk.nvariants is None else min(ncols, chunk.nvariants)

    if major == "variant":
        nrows, ncols = ncols, nrows
        row_chunk, col_chunk = col_chunk, row_chunk

    G = read_bed(fn, nrows, ncols, row_chunk, col_chunk, ref)
    if major == "sample":
        G = G.T

    return G


def _check_bed_header(fn):
    with open(fn, "rb") as f:
        arr = f.read(2)
        if len(arr) < 2:
            raise ValueError("Couldn't read BED header: %s." % fn)
        ok = arr[0] == 108 and arr[1] == 27
        if not ok:
            raise ValueError("Invalid BED file: %s." % fn)


def _major_order(fn):
    with open(fn, "rb") as f:
        f.seek(2)
        arr = f.read(1)
        if len(arr) < 1:
            raise ValueError("Couldn't read column order: %s." % fn)
        if arr[0] == 1:
            return "variant"
        elif arr[0] == 0:
            return "sample"
        msg = "Invalid matrix layout. Maybe it is a PLINK2 file?"
        msg += " PLINK2 is not supported yet."
        raise ValueError(msg)


def _clean_prefixes(prefixes):
    paths = []
    for p in prefixes:
        dirn = dirname(p)
        basen = basename(p)
        base = ".".join(basen.split(".")[:-1])
        if len(base) == 0:
            path = p
        else:
            ext = basen.split(".")[-1]
            if ext not in ["bed", "fam", "bim", "nosex", "log"]:
                base += "." + ext
            path = join(dirn, base)
        paths.append(path)
    return list(set(paths))



from deprecated.sphinx import deprecated


@deprecated(reason="use function :func:`get_data_folder` instead.", version="2.0.0")
def example_file_prefix():
    """
    Data files prefix.
    """
    import os

    p = __import__("pandas_plink").__path__[0]
    return os.path.join(p, "test", "data_files", "data")


def get_data_folder():
    """
    Path to the folder containing example files.
    """
    import os

    p = __import__("pandas_plink").__path__[0]
    return os.path.join(p, "test", "data_files")



from os.path import dirname, join, realpath

from numpy.testing import assert_almost_equal, assert_equal

from pandas_plink import read_rel


def test_read_rel():
    datafiles = join(dirname(realpath(__file__)), "data_files")

    filepath = join(datafiles, "rel", "plink.rel")
    K = read_rel(filepath)
    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.789741, -0.158418])
    assert_equal(K.sample_0[4].data, "NA12489")

    filepath = join(datafiles, "rel", "plink.rel")
    filepath_id = join(datafiles, "rel", "plink.rel.id")
    K = read_rel(filepath, filepath_id)
    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], [0.789741, -0.158418])
    assert_equal(K.sample_0[4].data, "NA12489")


def test_read_rel_bin():
    datafiles = join(dirname(realpath(__file__)), "data_files")
    desired = [0.8857815151328152, -0.09149171449203966]

    filepath = join(datafiles, "rel-bin", "plink2.rel.bin")
    K = read_rel(filepath)
    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], desired)
    assert_equal(K.sample_0[4].data, "NA12489")

    filepath = join(datafiles, "rel-bin", "plink2.rel.bin")
    filepath_id = join(datafiles, "rel-bin", "plink2.rel.id")
    K = read_rel(filepath, filepath_id)
    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], desired)
    assert_equal(K.sample_0[4].data, "NA12489")


def test_read_rel_zs():
    datafiles = join(dirname(realpath(__file__)), "data_files")
    desired = [0.885782, -0.0914917]

    filepath = join(datafiles, "rel-zs", "plink2.rel.zst")
    K = read_rel(filepath)
    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], desired)
    assert_equal(K.sample_0[4].data, "NA12489")

    filepath = join(datafiles, "rel-zs", "plink2.rel.zst")
    filepath_id = join(datafiles, "rel-zs", "plink2.rel.id")
    K = read_rel(filepath, filepath_id)
    assert_almost_equal([K.data[0, 0], K.data[-2, 5]], desired)
    assert_equal(K.sample_0[4].data, "NA12489")



from numpy import absolute, ascontiguousarray, empty, float32, nan, uint8, uint64, zeros

from ._allele import Allele

__all__ = ["read_bed"]


def read_bed(filepath, nrows, ncols, row_chunk, col_chunk, ref: Allele):
    from dask.array import concatenate, from_delayed
    from dask.delayed import delayed

    row_start = 0
    col_xs = []
    while row_start < nrows:
        row_end = min(row_start + row_chunk, nrows)
        col_start = 0
        row_xs = []
        while col_start < ncols:
            col_end = min(col_start + col_chunk, ncols)

            x = delayed(_read_bed_chunk, None, True, None, False)(
                filepath,
                nrows,
                ncols,
                row_start,
                row_end,
                col_start,
                col_end,
                ref,
            )

            shape = (row_end - row_start, col_end - col_start)
            row_xs.append(from_delayed(x, shape, float32))
            col_start = col_end

        col_xs.append(concatenate(row_xs, 1, True))
        row_start = row_end
    return concatenate(col_xs, 0, True)


def _read_bed_chunk(
    filepath, nrows, ncols, row_start, row_end, col_start, col_end, ref: Allele
):
    from .bed_reader import ffi, lib

    base_type = uint8
    base_size = base_type().nbytes
    base_repr = "uint8_t"

    X = zeros((row_end - row_start, col_end - col_start), base_type)
    assert X.flags.aligned

    strides = empty(2, uint64)
    strides[:] = X.strides
    strides //= base_size

    e = lib.read_bed_chunk(
        filepath.encode(),
        nrows,
        ncols,
        row_start,
        col_start,
        row_end,
        col_end,
        ffi.cast(f"{base_repr} *", X.ctypes.data),
        ffi.cast("uint64_t *", strides.ctypes.data),
    )
    if e != 0:
        raise RuntimeError(f"Failure while reading BED file {filepath}.")

    X = ascontiguousarray(X, float32)
    X[X == 3] = nan
    if ref == Allele.a0:
        X -= 2
        absolute(X, out=X)

    return X
